#!/bin/sh
#\
exec tclsh "$0" 

namespace eval ::tk_docapp \
{
  namespace export Main
}


# we load the msgcat package in a global scope  
package require msgcat

# we check if a language is set from the operation system
if {![info exist env(LANG)]} {set env(LANG) "en"}
msgcat::mclocale $env(LANG)
msgcat::mcload [file join [pwd] msgs]

# --------------------------------
# the mother of all procedures :o)
# --------------------------------
proc ::tk_docapp::Main {} \
{
  variable CONFIG; variable VAR

  ::tk_docapp::Startup

  ::tk_docapp::MainWindow

  ::tk_docapp::CallBacks

  ::tk_docapp::ShowMainWindow 
}


# ---------------------------------------------
# set the default values
# ---------------------------------------------
proc ::tk_docapp::SetDefaultValues {} \
{
  variable DEFAULTS; variable VAR;

  set DEFAULTS(Geometry) "90x210+20+20"
  set DEFAULTS(Beep) 0
  set DEFAULTS(AvgDischargeCurrent) 1.0

  set VAR(StopBeep) 0
  set VAR(RefreshId) ""
  set VAR(DateShowed) 0
  set VAR(LastTime) ""
}


# ---------------------------------------------
# set the defaults and load all needed packages
# ---------------------------------------------
proc ::tk_docapp::Startup {} \
{
  variable ABOUT; variable VAR; variable CONFIGFILE
  variable DEFAULTS; variable CONFIG;

  global env auto_path argc argv

  package require Tk
  package require Ttk

  set VAR(NameSpace) [namespace current]
  set VAR(StartDir) [pwd]
  # the return value is set to error as default
  set VAR(ReturnValue) 1

  #-----the informations for the about boxes and the window border-----
  set AboutFile [file join $VAR(StartDir) "about.tcl"]

  if {[file exists $AboutFile]} \
  {
    source $AboutFile
  } else \
  {
    tk_messageBox -message [msgcat::mc "I can't find the file:\n\n\"%s\"\n\nPlease install the program again." $AboutFile]\
    -title [msgcat::mc "File not found"] -icon error -type ok
    destroy .
    exit
  }

  # we check the existens of lib directory
  set VAR(LibDir) [file join $VAR(StartDir) lib]

  if { [file exists $VAR(LibDir)] !="1" } \
  {
    tk_messageBox -message [msgcat::mc "I can't find the program libraries!\n\nPlease install the program again."]\
    -title [msgcat::mc "Where are my libraries?"] -icon error -type ok
    destroy .
    exit
  } else \
  {
    lappend auto_path $VAR(LibDir)
  }

  # set the default values
  ::tk_docapp::SetDefaultValues

  # we load all needed packages
  ::tk_docapp::Packageload [list FR_config FR_messageBox]

  # we read and parse the configuration file
  set ConfigFileDir [file join $env(HOME) .${ABOUT(Name)}]

  if {![file exists $ConfigFileDir]} \
  {
    catch {file mkdir $ConfigFileDir}
  }

  set CONFIGFILE [file join $ConfigFileDir .${ABOUT(Name)}_conf.tcl]

  ::FR_config::ReadConfig $VAR(NameSpace)
}


# ----------------------------------------
# the package loader
# we load the needed package only one time
# ----------------------------------------
proc ::tk_docapp::Packageload {Packages} \
{
  variable VAR

  foreach Arg $Packages \
  {
    if {![info exists VAR($Arg)]} \
    {
      if { [catch {package require $Arg} output] !="0" } \
      {
	tk_messageBox -message [msgcat::mc "I can't load this part from the program library:\n\n%s\n\nPlease install the program again." $Arg]\
	-title [msgcat::mc "Parts of the program couldn't be loaded"] -icon error -type ok
	destroy .
	exit
      } else \
      {
        set VAR($Arg) 1
      }
    }
  }
}


# ------------------------------------------------------------------
# creates the main window and starts the creation of the sub windows
# ------------------------------------------------------------------
proc ::tk_docapp::MainWindow {} \
{
  variable ABOUT; variable WIN

  wm title . [msgcat::mc "%s version %s from %s%s" $ABOUT(Name) $ABOUT(Version) $ABOUT(Date) $ABOUT(Year)]
  wm iconname . $ABOUT(Name)
  wm protocol . WM_DELETE_WINDOW ::tk_docapp::Exit
  wm minsize . 20 20
  wm withdraw .

  # the frame for the menu line
  ttk::frame .menuFrame
  pack .menuFrame -fill x -side top -padx 5 -pady 5

  # the menu line on top of the main window
  ::tk_docapp::CreateMenuLine .menuFrame

  # the frame for the clock
  ttk::labelframe .clockFrame -text [msgcat::mc "Time"]
  set WIN(TimeLabelFrame) ".clockFrame"
  pack $WIN(TimeLabelFrame) -side top -fill x -padx 5 -pady 5

  # show the time with a label
  ttk::label $WIN(TimeLabelFrame).timelabel -anchor center
  set WIN(TimeLabel) "$WIN(TimeLabelFrame).timelabel"
  pack $WIN(TimeLabel) -side top -pady {0 5}

  # frame for the battery status
  ttk::labelframe .batteryFrame -text [msgcat::mc "Battery"]
  pack .batteryFrame -side top -fill x -padx 5 -pady 5

  # battery status in procent and left time
  ttk::label .batteryFrame.batteryPercent -anchor center
  set WIN(BatteryPercent) ".batteryFrame.batteryPercent"
  pack $WIN(BatteryPercent) -side top -pady {0 5}

  ttk::label .batteryFrame.batteryLeftTime -anchor center
  set WIN(BatteryLeftTime) ".batteryFrame.batteryLeftTime"
  pack $WIN(BatteryLeftTime) -side top -pady {0 5}

  ttk::label .batteryFrame.batteryStatus -anchor center
  set WIN(BatteryStatus) ".batteryFrame.batteryStatus"
  pack $WIN(BatteryStatus) -side top -pady {0 5}

  frame .stopBeepFrame
  pack .stopBeepFrame -side top -fill x -padx 5 -pady 5

  ttk::button .stopBeepFrame.stopBeepButton -text [msgcat::mc "Stop beep"]
  set WIN(StopBeepButton) ".stopBeepFrame.stopBeepButton"
  pack $WIN(StopBeepButton) -side top
}


# ----------------------
# configures the widgets
# ----------------------
proc ::tk_docapp::CallBacks {} \
{
  variable WIN; variable CONFIG
  variable VAR

  # the stop beep button
  $WIN(StopBeepButton) configure -command {set ::tk_docapp::VAR(StopBeep) 1}

  # the binding for the time label (switch to display the date)
  bind $WIN(TimeLabel) <Enter> {::tk_docapp::ShowDate}
  bind $WIN(TimeLabel) <Leave> {::tk_docapp::RestoreTime}

  # we call a refresh function
  ::tk_docapp::Refresh
}


# ---------------------------------------
# shows the main window in the right size
# ---------------------------------------
proc ::tk_docapp::ShowMainWindow {} \
{
  variable CONFIG

  wm geometry . $CONFIG(Geometry)
  wm deiconify .
  update
}


# ---------------------------------------------
# compose the text for the about window
# ---------------------------------------------
proc ::tk_docapp::GenerateAboutText {} \
{
  variable ABOUT;

  set Text $ABOUT(Header)

  append Text "\n\n" [msgcat::mc "Version: %s from %s%s" $ABOUT(Version) $ABOUT(Date) $ABOUT(Year)]
  append Text "\n\n" $ABOUT(License)

  return $Text
}


# ---------------------------------------------
# show the about window
# ---------------------------------------------
proc ::tk_docapp::ShowAboutText {} \
{
  ::FR_messageBox::Create -title [msgcat::mc "About"] -message [::tk_docapp::GenerateAboutText]
}


# ----------------------------------------
# creates the menu line
# ----------------------------------------
proc ::tk_docapp::CreateMenuLine {MenuWindow} \
{
  variable CONFIG

  # the file menu
  menubutton ${MenuWindow}.file -text [msgcat::mc "File"] -menu ${MenuWindow}.file.sub -underline 0
  menu ${MenuWindow}.file.sub -tearoff 0
  menu ${MenuWindow}.file.sub.beep -tearoff 0

  ${MenuWindow}.file.sub add cascade -label [msgcat::mc "Beep on low battery"] -menu [list ${MenuWindow}.file.sub.beep]
  ${MenuWindow}.file.sub add separator
  ${MenuWindow}.file.sub add command -label [msgcat::mc "Save settings / position"] -accelerator "Ctr+s" \
  -command {::FR_config::SaveConfig $::tk_docapp::VAR(NameSpace)}
  ${MenuWindow}.file.sub add separator
  ${MenuWindow}.file.sub add command -label [msgcat::mc "Exit"] -accelerator "Ctr+q" -command {::tk_docapp::Exit}

  ${MenuWindow}.file.sub.beep add radiobutton -label [msgcat::mc "Never"] -variable ::tk_docapp::CONFIG(Beep) -value 0
  ${MenuWindow}.file.sub.beep add radiobutton -label [msgcat::mc "At 5% capacity"] -variable ::tk_docapp::CONFIG(Beep) -value 5
  ${MenuWindow}.file.sub.beep add radiobutton -label [msgcat::mc "At 10% capacity"] -variable ::tk_docapp::CONFIG(Beep) -value 10

  bind . <Control-s> [list ${MenuWindow}.file.sub invoke 2]
  bind ${MenuWindow}.file.sub <Control-s> [list ::FR_general::MenuBinding ${MenuWindow}.file.sub 2]
  bind . <Control-q> [list ${MenuWindow}.file.sub invoke 4]
  bind ${MenuWindow}.file.sub <Control-q> [list ::FR_general::MenuBinding ${MenuWindow}.file.sub 4] 

  menubutton ${MenuWindow}.help -text [msgcat::mc "Help"] -menu ${MenuWindow}.help.sub -underline 0
  menu ${MenuWindow}.help.sub -tearoff 0 

  ${MenuWindow}.help.sub add command -label [msgcat::mc "About"] -accelerator "Alt+a" -command {::tk_docapp::ShowAboutText} 

  bind . <Alt-a>  [list ${MenuWindow}.help.sub invoke 0]
  bind ${MenuWindow}.help.sub <Alt-a> [list ::FR_general::MenuBinding ${MenuWindow}.help.sub 0]

  pack ${MenuWindow}.file -side left
  pack ${MenuWindow}.help -side right
}


# ------------------------------------------------
# we save the configuration and finish the program
# ------------------------------------------------
proc ::tk_docapp::Exit {} \
{
  variable CONFIG; variable VAR

  # we stop the last refresh call
  after cancel $VAR(RefreshId)

  ::FR_config::SaveConfig $VAR(NameSpace)

  catch {puts [msgcat::mc "Goodby!"]}

  destroy .

  exit $VAR(ReturnValue)
}


# ------------------------------------------------
# the refresh function called each minute
# ------------------------------------------------
proc ::tk_docapp::Refresh {} \
{
  variable VAR; variable WIN;
  global tcl_platform


  # we delete an other refresh command
  after cancel $VAR(RefreshId)

  # first we refresh the time
  set VAR(LastTime) [::tk_docapp::GetTime]

  # we update the time information only if the date isn't showed
  if {$VAR(DateShowed) == 0} \
  {
    $WIN(TimeLabel) configure -text $VAR(LastTime)
  }

  # we fetch the battery and ac adapter status
  #  (stored for further use) "NetBSD"  {set BatStatus [::tk_docapp::Battery_NetBSD]}
  switch -glob -- $tcl_platform(os) \
  {
    "OpenBSD" {set BatStatus [::tk_docapp::Battery_OpenBSD]} \
    *         {set BatStatus 1} 
  }

  # we check the if the call of the background tools was successful
  if {[lindex $BatStatus 0] == 0} \
  {
    $WIN(BatteryPercent) configure -text [lindex $BatStatus 1]
    $WIN(BatteryLeftTime) configure -text [lindex $BatStatus 2]
    $WIN(BatteryStatus) configure -text [lindex $BatStatus 3]

    set BeepReturn 0

    # Beep only if the battery is discharged
    if {[lindex $BatStatus 3] eq [msgcat::mc "Discharging"]} \
    {
      set BeepReturn [::tk_docapp::Beep [lindex $BatStatus 1]]
    }

    # if we beep we display the battery status red
    if {($BeepReturn == 1) && ([$WIN(BatteryPercent) cget -foreground] eq "")} \
    {
      foreach Arg {$WIN(BatteryPercent) $WIN(BatteryLeftTime) $WIN(BatteryStatus)} \
      {
        $Arg configure -foreground red
      }
    }

    if {($BeepReturn == 0) && ([$WIN(BatteryPercent) cget -foreground] eq "red")} \
    {
      foreach Arg {$WIN(BatteryPercent) $WIN(BatteryLeftTime) $WIN(BatteryStatus)} \
      {
	$Arg configure -foreground {}
      }
    }
  } \
  else \
  {
    $WIN(BatteryPercent) configure -text "0%"
    $WIN(BatteryLeftTime) configure -text "0:00"
    $WIN(BatteryStatus) configure -text [msgcat::mc "Unknown"] 
  }

  # the next refresh is in one minute
  set VAR(RefreshId)  [after 60000 ::tk_docapp::Refresh] 
}


# ------------------------------------------------
# the time get / formating proc
# ------------------------------------------------
proc ::tk_docapp::GetTime {} \
{
  # time as integer seconds
  set Time [clock seconds]

  # format it to a more readable form... :o)
  set Time [clock format $Time -format "%H:%M"]

  return $Time 
}


# ------------------------------------------------
# the date get / formating proc
# ------------------------------------------------
proc ::tk_docapp::GetDate {} \
{
  # time as integer seconds
  set Time [clock seconds]

  # format it to a more readable form... :o)
  #set Date [clock format $Time -format "%d.%B.%Y"]
  set Date [clock format $Time -format "%Ex"]

  return $Date 
}


# ------------------------------------------------
# ask sysctl for the battery status under OpenBSD
# ------------------------------------------------ 
proc ::tk_docapp::Battery_OpenBSD {} \
{
  variable CONFIG; variable DEFAULTS;
  
  set Error 1

  # call sysctl for all hardware sensors information
  set ReturnVal [catch {exec sysctl hw.sensors.acpibat0} SysCtlOutput]

  ######################
  # for debugging only #
  ######################
  # puts $SysCtlOutput

  # if the call fails we return the default values
  if {$ReturnVal} {return $Error}

  # generate a list with one element per line
  set SysCtlOutput [split $SysCtlOutput "\n"] 

  # we look for the full capacity
  ###############################
  set Index 0

  foreach Line $SysCtlOutput \
  {
    set ReturnVal [scan $Line "hw.sensors.acpibat0.amphour0=%f" BatCapFull]

    # we found the line
    if {$ReturnVal > 0} \
    {
      # we can delete the list element to save time in the further scans
      set SysCtlOutput [lreplace $SysCtlOutput $Index $Index]
      # we can left the loop
      break
    }

    incr Index
  }


  # the last scan wasn't succesful -> we didn't find anything
  if {$ReturnVal < 1} {return $Error}

  # we look for the current capacity
  ##################################
  set Index 0

  foreach Line $SysCtlOutput \
  {
    set ReturnVal [scan $Line "hw.sensors.acpibat0.amphour3=%f" BatCapCurrent]

    if {$ReturnVal > 0} \
    {
      # we can delete the list element to save time in the further scans
      set SysCtlOutput [lreplace $SysCtlOutput $Index $Index]
      # we can left the loop
      break
    }

    incr Index
  }

  # the last scan wasn't succesful -> we didn't find anything
  if {$ReturnVal < 1} {return $Error}

  # we look for the consumed current
  ##################################
  set Index 0

  foreach Line $SysCtlOutput \
  {
    set ReturnVal [scan $Line "hw.sensors.acpibat0.current0=%f" BatCurrent]

    if {$ReturnVal > 0} \
    {
      # we can delete the list element to save time in the further scans
      set SysCtlOutput [lreplace $SysCtlOutput $Index $Index]
      # we can left the loop
      break
    }

    incr Index
  }

  # the last scan wasn't succesful -> we didn't find anything
  if {$ReturnVal < 1} {return $Error}

  # we look for the battery status
  ################################ 
  foreach Line $SysCtlOutput \
  {
    set ReturnVal [scan $Line "hw.sensors.acpibat0.raw0=%d" BatStat]

    if {$ReturnVal > 0} \
    {
      # we can left the loop
      break
    }
  }

  # the last scan wasn't succesful -> we didn't find anything
  if {$ReturnVal == -1} {return $Error}

  # calculate the battery capacity 
  ################################ 
  if {$BatCapFull == 0} {return $Error}; # protection for the division
  
  set BatCapLeft [expr {$BatCapCurrent / $BatCapFull * 100.0}]
  
  # the last cap full can be smaller the the current cap full
  # we limited the value to 100% 
  if {$BatCapLeft > 100.0} {set BatCapLeft 100.0}
 
  set BatCapLeft [format "%d%%" [expr {int($BatCapLeft)}]]

  # calculate the average of the discharge current
  ################################################
  # the consumed current is only true if the battery is discharged
  # and not full
  if {($BatStat == 1) && ($BatCapLeft ne "100%")} \
  {
    # the average is 90% from the old value and 10% from the current value
    set CONFIG(AvgDischargeCurrent) [expr {0.9 * $CONFIG(AvgDischargeCurrent) + 0.1 * $BatCurrent}]
  
    # protection for the following division
    if {$CONFIG(AvgDischargeCurrent) == 0} {set CONFIG(AvgDischargeCurrent) $DEFAULTS(AvgDischargeCurrent)}
  }

  # calculate the left battery working time
  #########################################
  
  # the battery time in hours
  set Hours [expr {$BatCapCurrent / $CONFIG(AvgDischargeCurrent)}]
  
  # we separate the minutes from the full hours
  set Minutes [expr {$Hours - floor($Hours)}]
  
  # Minutes in the 60 minutes format
  set Minutes [expr {$Minutes * 60.0}]
  
  # bring the time information in the form e.g. 2:03
  set BatTimeLeft [format "%d:%02d" [expr {int($Hours)}] [expr {int($Minutes)}]]

  # set the battery status
  ########################
  switch $BatStat \
  {
    "0" {set BatStatus [msgcat::mc "Idle"]} \
    "1" {set BatStatus [msgcat::mc "Discharging"]} \
    "2" {set BatStatus [msgcat::mc "Charging"]} \
    *   {set BatStatus [msgcat::mc "Unknown"]}
  }

  if {$BatCapLeft eq "100%"} \
  {
    set BatStatus [msgcat::mc "Full"]
  }

  set Error 0

  return [list $Error $BatCapLeft $BatTimeLeft $BatStatus]
}


# ------------------------------------------------
# show the date information insteed of the time
# ------------------------------------------------
proc ::tk_docapp::ShowDate {} \
{
  variable VAR; variable WIN

  # this avoids to set the time during displaying the date
  set VAR(DateShowed) 1


  $WIN(TimeLabelFrame) configure -text [msgcat::mc "Date"]
  $WIN(TimeLabel) configure -text [::tk_docapp::GetDate]
}


# ------------------------------------------------
# restore the time information after displaying
# the date
# ------------------------------------------------
proc ::tk_docapp::RestoreTime {} \
{
  variable VAR; variable WIN

  # now we can display the time again
  set VAR(DateShowed) 0

  $WIN(TimeLabelFrame) configure -text [msgcat::mc "Time"]
  $WIN(TimeLabel) configure -text $VAR(LastTime) 
}


# ------------------------------------------------
# Beep if the battery capacity is too low
# ------------------------------------------------
proc ::tk_docapp::Beep {BatCapLeft} \
{
  variable VAR; variable CONFIG

  # we return immediately if now beep is wished
  if {($CONFIG(Beep) == 0) || ($VAR(StopBeep) == 1)} {return}

  # remove the tailing "%"
  set BatCapLeft [string trimright $BatCapLeft "%"]
  
  # beep at lower than 5% or 10%
  if {($CONFIG(Beep) == 5) && ($BatCapLeft <= $CONFIG(Beep))} \
  {
    set Beep 1
  } \
  elseif {($CONFIG(Beep) == 10) && ($BatCapLeft <= $CONFIG(Beep))} \
  {
    set Beep 1
  } \
  else \
  {
    set Beep 0
  }

  if {$Beep} \
  { 
    puts [msgcat::mc "Only %s%% battery capacity left!" $BatCapLeft]
    bell
    after 500 {bell}
  }

  return $Beep
}


# -----------------
# start the program
# -----------------
::tk_docapp::Main

